@@ -13,15 +13,22 @@ use Devel::Peek;
 use Encode qw(encode_utf8 is_utf8);
 use Scalar::Util qw(reftype blessed refaddr);
 use Config;
+use Carp qw(confess);
 
 # Dynamically load constants from whatever is being tested
 our ($Class, $ConstClass);
 BEGIN {
-    if (-e "lib/Sereal/Encoder") {
+    if (-e "lib/Sereal/Encoder.pm") {
         $Class = 'Sereal::Encoder';
     }
-    elsif (-e "lib/Sereal/Decoder") {
+    elsif (-e "lib/Sereal/Decoder.pm") {
         $Class = 'Sereal::Decoder';
+    }
+    elsif (-e "lib/Sereal/Merger.pm") {
+        $Class = 'Sereal::Merger';
+    }
+    elsif (-e "lib/Sereal/Splitter.pm") {
+        $Class = 'Sereal::Splitter';
     } else {
         die "Could not find an applicable Sereal constants location";
     }
@@ -115,10 +122,21 @@ sub dump_bless {
 }
 
 sub short_string {
-    die if length($_[0]) > SRL_MASK_SHORT_BINARY_LEN;
-    my $tag = SRL_HDR_SHORT_BINARY_LOW + length($_[0]);
-    $tag |= SRL_HDR_TRACK_FLAG if $_[1];
-    return pack("c a*",$tag,$_[0]);
+    my ($str, $alias)= @_;
+    $alias ||= 0;
+    my $length= length($str);
+    if ($length > SRL_MASK_SHORT_BINARY_LEN) {
+        confess "String too long for short_string(), alias=$alias length=$length";
+    }
+    my $tag = SRL_HDR_SHORT_BINARY_LOW + length($str);
+    if ($tag > SRL_HDR_SHORT_BINARY_HIGH) {
+        confess "Tag value larger than SRL_HDR_SHORT_BINARY_HIGH, tag=$tag; alias=$alias; length=$length";
+    }
+    $tag |= SRL_HDR_TRACK_FLAG if $alias;
+    if ($tag > 255) {
+        confess "Tag value over 255 in short_string(), tag=$tag; alias=$alias; length=$length; SRL_HDR_TRACK_FLAG=", SRL_HDR_TRACK_FLAG;
+    }
+    return chr($tag) . $str;
 }
 
 sub integer {
@@ -180,6 +198,17 @@ sub offseti {
     }
 }
 
+sub _permute {
+    return [] unless @_;
+    my $vals= shift;
+    my @rest= _permute(@_);
+    map { my $v= $_; map { [ $v, @$_ ] } @rest } @$vals;
+}
+
+sub permute_array {
+    map { array(@$_) }  _permute(@_);
+}
+
 sub debug_checks {
     my ($data_ref, $encoded_ref, $decoded_ref, $debug) = @_;
     if ($debug or defined $ENV{DEBUG_SEREAL}) {
@@ -234,14 +263,13 @@ sub setup_tests {
         [[], array(), "empty array ref"],
         [[1,2,3], array(chr(0b0000_0001), chr(0b0000_0010), chr(0b0000_0011)), "array ref"],
         [1000, chr(SRL_HDR_VARINT).varint(1000), "large int"],
-        [ [1..1000],
+        [ [ map { $_, undef } 1..1000 ],
             array(
-                (map chr, (1 .. SRL_POS_MAX_SIZE)),
-                (map chr(SRL_HDR_VARINT) . varint($_), ((SRL_POS_MAX_SIZE+1) .. 1000))
+                (map { chr($_) => chr(SRL_HDR_UNDEF) } (1 .. SRL_POS_MAX_SIZE)),
+                (map { chr(SRL_HDR_VARINT) . varint($_) => chr(SRL_HDR_UNDEF) } ((SRL_POS_MAX_SIZE+1) .. 1000))
             ),
-            "array ref with pos and varints"
+            "array ref with pos and varints and undef"
         ],
-
         [{}, hash(), "empty hash ref"],
         [{foo => "baaaaar"}, hash(short_string("foo"),short_string("baaaaar")), "simple hash ref"],
         [
@@ -540,17 +568,33 @@ sub setup_tests {
             sub { \@_ }->(!1,!0),
             array(chr(SRL_HDR_FALSE),chr(SRL_HDR_TRUE)),  # this is the "correct" response.
             "true/false (prefered order)",
-            array(chr(SRL_HDR_FALSE),short_string("1")),  # this is what threaded perls will probably match
-            array(short_string(""),chr(SRL_HDR_TRUE)),    # accept this also (but we dont expect we will)
-            array(short_string(""),short_string("1")),    # accept this also (but we dont expect we will)
+            permute_array(
+                [
+                    short_string(""),
+                    chr(SRL_HDR_FALSE),
+                ],
+                [
+                    chr(SRL_HDR_TRUE),
+                    short_string("1"),
+                    integer(1)
+                ]
+            ),  # this is what threaded perls will probably match
         ],
         [
             sub { \@_ }->(!1,!0),
             array(short_string(""),short_string("1")),    # this is the expected value on perl 5.14 unthreaded
             "true/false (reversed alternates)",
-            array(short_string(""),chr(SRL_HDR_TRUE)),    # from here we just reverse the order from the first test
-            array(chr(SRL_HDR_FALSE),short_string("1")),  # ....
-            array(chr(SRL_HDR_FALSE),chr(SRL_HDR_TRUE)),
+            permute_array(
+                [
+                    short_string(""),
+                    chr(SRL_HDR_FALSE)
+                ],
+                [
+                    chr(SRL_HDR_TRUE),
+                    integer(1),
+                    short_string("1")
+                ]
+            ),
         ],
     );
 }
@@ -572,7 +616,6 @@ sub have_encoder_and_decoder {
     # $Class is the already-loaded class, so the one we're testing
     my $need = $Class =~ /Encoder/ ? "Decoder" : "Encoder";
     my $need_class = "Sereal::$need";
-    my %compat_versions = map {$_ => 1} $Class->_test_compat();
 
     if (defined(my $top_dir = get_git_top_dir())) {
         my $blib_dir = File::Spec->catdir($top_dir, 'Perl', $need, "blib");
@@ -581,6 +624,11 @@ sub have_encoder_and_decoder {
             blib->import($blib_dir);
         }
     }
+    eval "use $Class; 1"
+    or do {
+        note("Could not locate $Class for testing" . ($@ ? " (Exception: $@)" : ""));
+        return();
+    };
 
     eval "use $need_class; 1"
     or do {
@@ -595,6 +643,7 @@ sub have_encoder_and_decoder {
     }
     $cmp_v =~ s/_//;
     $cmp_v = sprintf("%.2f", int($cmp_v*100)/100);
+    my %compat_versions = map {$_ => 1} $Class->_test_compat();
     if (not defined $cmp_v or not exists $compat_versions{$cmp_v}) {
         note("Could not load correct version of $need_class for testing "
              ."(got: $cmp_v, needed any of ".join(", ", keys %compat_versions).")");
@@ -621,6 +670,10 @@ my $lots_of_9C = do {
 my $max_iv = ~0 >> 1;
 my $min_iv = do {use integer; -$max_iv-1}; # 2s complement assumption
 
+my @numstr= map { ; no warnings; $_ < 0 and warn "this shouldnt happpen"; $_ }
+    ( "    1    ", "0.0", "00000.0000", "0.0.0.0", ".0","    .0", " 22",
+      "01", "01.1", "   0   ", ".0", "0.001", ".1", "  .1", ".2", "00", ".00",
+      "0 but true", "0E0");
 my $eng0e0= "0e0";
 my $eng0e1= "0e1";
 my $eng2= "1e3";
@@ -638,7 +691,7 @@ our @ScalarRoundtripTests = (
     ["largeish negative int -12345678", -12345678],
 
     (
-        map {["integer: $_", $_]} (
+        map {["integer: $_", 0+$_]} (
             # IV bounds of 8 bits
             -1, 0, 1, -127, -128, -129, 42, 126, 127, 128, 129, 254, 255, 256, 257,
             # IV bounds of 32 bits
@@ -653,8 +706,7 @@ our @ScalarRoundtripTests = (
             $eng0e0, $eng0e1, $eng2,
         )
     ),
-
-    ["float", 0.2],
+    (map { ["float $_", 0+$_] } (0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9)),
     ["short ascii string", "fooo"],
     ["short latin1 string", "Müller"],
     ["short utf8 string", do {use utf8; " עדיין ח"} ],
@@ -679,7 +731,7 @@ our @ScalarRoundtripTests = (
         1023,1024,1025,
         8191,8192,8193,
     )),
-
+    ["troublesome num/strs", @numstr],
     ["long latin1 string", "üll" x 10000],
     ["long utf8 string", do {use utf8; " עדיין חשב" x 10000}],
     ["long utf8 string with only ascii", do {use utf8; "foo" x 10000}],
@@ -690,8 +742,8 @@ our @ScalarRoundtripTests = (
     ["regexp with modifiers", qr/foo/i],
     ["float", 123013.139],
     ["negative float",-1234.59],
-    ["small float",0.41],
-    ["negative small float",-0.13],
+    ["small float 0.41",0.41],
+    ["negative small float -0.13",-0.13],
     ["small int", 123],
     ["empty string", ''],
     ["simple array", []],
@@ -703,7 +755,16 @@ our @ScalarRoundtripTests = (
     ["deep nest", [1,2,[3,4,{5=>6,7=>{8=>[]},9=>{}},{},[]]]],
     ["complex hash", {
         foo => 123,
-        bar => -159.23 ,
+        bar => -159, pi => 3,
+        'baz' =>"foo",
+        'bop \''=> "\10"
+        ,'bop \'\\'=> "\x{100}" ,
+        'bop \'x\\x'    =>"x\x{100}"   , 'bing' =>   "x\x{100}",
+        x=>'y', z => 'p', i=> '1', l=>" \10", m=>"\10 ", n => " \10 ",
+    }],
+    ["complex hash with float", {
+        foo => 123,
+        bar => -159.23, a_pi => 3.14159,
         'baz' =>"foo",
         'bop \''=> "\10"
         ,'bop \'\\'=> "\x{100}" ,
@@ -712,7 +773,16 @@ our @ScalarRoundtripTests = (
     }],
     ["more complex", {
         foo => [123],
-        "bar" => [-159.23 , { 'baz' => "foo", }, ],
+        "bar" => [-159, n => 3, { 'baz' => "foo", }, ],
+        'bop \''=> { "\10" => { 'bop \'\\'=> "\x{100}", h=>{
+        'bop \'x\\x'    =>"x\x{100}"   , 'bing' =>   "x\x{100}",
+        x=>'y',}, z => 'p' ,   }   ,
+        i    =>  '1' ,}, l=>" \10", m=>"\10 ", n => " \10 ",
+        o => undef ,p=>undef, q=>\undef, r=>\$eng0e0, u => \$eng0e1, w=>\$eng2
+    }],
+    ["more complex with float", {
+        foo => [123],
+        "bar" => [-159.23, a_pi => 3.14159, { 'baz' => "foo", }, ],
         'bop \''=> { "\10" => { 'bop \'\\'=> "\x{100}", h=>{
         'bop \'x\\x'    =>"x\x{100}"   , 'bing' =>   "x\x{100}",
         x=>'y',}, z => 'p' ,   }   ,
@@ -804,6 +874,7 @@ sub _test {
     }
     return;
 }
+
 sub _cmp_str {
     my ($v1, $v2)= @_;
     my $v1_is_utf8= is_utf8($v1);
@@ -819,7 +890,6 @@ sub _cmp_str {
                     and $diff_start < length($v2)
                     and substr($v1, $diff_start,1) eq substr($v2, $diff_start,1);
     my $diff_end= length($v1) < length($v2) ? length($v1) : length($v2);
-    $diff_end-- if $diff_end;
 
     $diff_end-- while $diff_end > $diff_start
                   and $diff_end > $diff_start
@@ -866,8 +936,8 @@ sub _cmp_str {
 
     my $ret= sprintf(  "strings different\n"
                      . "first string difference at octet offset %d%s\n"
-                     . "want-octets = %*s (octets: %*d, utf8-flag: %d)\n"
                      . " got-octets = %*s (octets: %*d, utf8-flag: %d)\n"
+                     . "want-octets = %*s (octets: %*d, utf8-flag: %d)\n"
         ,$diff_start, $issues,
         -$pad, $q1, $lpad, length($v1), $v1_is_utf8,
         -$pad, $q2, $lpad, length($v2), $v2_is_utf8,
@@ -953,6 +1023,7 @@ sub run_roundtrip_tests_internal {
     foreach my $rt (@RoundtripTests) {
         my ($name, $data) = @$rt;
 
+        TODO:
         foreach my $meth (
               ['object-oriented',
                 sub {$encoder->encode($_[0])},
@@ -972,6 +1043,8 @@ sub run_roundtrip_tests_internal {
         ) {
             my ($mname, $enc, $dec) = @$meth;
 
+            local $TODO= $name=~/TODO/ ? $name : undef;
+
             next if $mname =~ /header/ and $opt->{use_protocol_v1};
 
             my $encoded;
@@ -1066,10 +1139,14 @@ sub run_roundtrip_tests_internal {
             deep_cmp($decoded3, $decoded,   "$name ($ename, $mname, decoded3 vs decoded)") or next; #test
             deep_cmp($decoded3, $decoded2,  "$name ($ename, $mname, decoded3 vs decoded2)") or next; #test
 
-            if ( $ename =~ /canon/) {
-                deep_cmp($encoded2, $encoded,  "$name ($ename, $mname, encoded2 vs encoded)") or next; #test
-                deep_cmp($encoded3, $encoded,  "$name ($ename, $mname, encoded3 vs encoded)") or next; #test
+            if ( $ename =~ /canon/ ) {
+                deep_cmp($encoded2, $encoded,  "$name ($ename, $mname, encoded2 vs encoded)") or do {
+                    diag Dumper($encoded2);
+                    diag Dumper($encoded);
+                    next; #test
+                };
                 deep_cmp($encoded3, $encoded2, "$name ($ename, $mname, encoded3 vs encoded2)") or next; #test
+                deep_cmp($encoded3, $encoded,  "$name ($ename, $mname, encoded3 vs encoded)") or next; #test
 
                 if ($ENV{SEREAL_TEST_SAVE_OUTPUT} and $mname eq 'object-oriented') {
                     use File::Path;
